@@ -2,7 +2,7 @@
  * Copyright 2005-2007 LAMP/EPFL
  * @author Martin Odersky
  */
-// $Id: Mixin.scala 12748 2007-08-31 15:33:03Z dragos $
+// $Id: Mixin.scala 14106 2008-02-22 13:24:21Z odersky $
 
 package scala.tools.nsc.transform
 
@@ -43,7 +43,7 @@ abstract class Mixin extends InfoTransform {
    */
   private def isImplementedStatically(sym: Symbol) =
     sym.owner.isImplClass && sym.isMethod &&
-    (!sym.isModule || sym.hasFlag(PRIVATE)) &&
+    (!sym.isModule || sym.hasFlag(PRIVATE | LIFTED)) &&
     (!(sym hasFlag (ACCESSOR | SUPERACCESSOR)) || sym.hasFlag(LAZY))
 
   /** A member of a trait is static only if it belongs only to the
@@ -59,7 +59,7 @@ abstract class Mixin extends InfoTransform {
    *  inherits the trait.
    */
   private def isForwarded(sym: Symbol) =
-    isImplementedStatically(sym) && !sym.isImplOnly && !sym.isCaseFactory
+    isImplementedStatically(sym) && !sym.isImplOnly
 
   /** Maps the type of an implementation class to its interface;
    *  maps all other types to themselves.
@@ -103,7 +103,7 @@ abstract class Mixin extends InfoTransform {
         if (settings.debug.value) {
           val other = bcs.head.info.nonPrivateDecl(member.name);
           log("rebindsuper " + bcs.head + " " + other + " " + other.tpe +
-              " " + other.hasFlag(DEFERRED))
+              " " + other.isDeferred)
         }
         sym = member.overridingSymbol(bcs.head).suchThat(sym => !sym.hasFlag(DEFERRED | BRIDGE))
         bcs = bcs.tail
@@ -155,7 +155,7 @@ abstract class Mixin extends InfoTransform {
 
       for (val member <- impl.info.decls.toList) {
         if (!member.isMethod && !member.isModule && !member.isModuleVar) {
-          assert(member.isTerm && !member.hasFlag(DEFERRED), member)
+          assert(member.isTerm && !member.isDeferred, member)
           if (member.getter(impl) hasFlag PRIVATE) {
             member.makeNotPrivate(clazz) // this will also make getter&setter not private
           }
@@ -221,32 +221,46 @@ abstract class Mixin extends InfoTransform {
        */
       def mixinTraitMembers(mixinClass: Symbol) {
         // For all members of a trait's interface do:
+        def isConcreteAccessor(member: Symbol) = 
+          (member hasFlag ACCESSOR) && 
+          (!(member hasFlag DEFERRED) || (member hasFlag lateDEFERRED))
+        def isOverridden(member: Symbol) =
+          isConcreteAccessor(member.overridingSymbol(clazz))
         for (val member <- mixinClass.info.decls.toList) {
-          if ((member hasFlag ACCESSOR) && 
-              (!(member hasFlag DEFERRED) || (member hasFlag lateDEFERRED))) { 
-            // mixin field accessors
-            val member1 = addMember(
-              clazz,
-              member.cloneSymbol(clazz)
-                setPos clazz.pos
-                setFlag FINAL resetFlag (DEFERRED | lateDEFERRED))
-            if (member.hasFlag(LAZY)) {
-              var init = implClass(mixinClass).info.decl(member.name)
-              assert(init != NoSymbol, "Could not find initializer for " + member.name)
-              initializer(member1) = init
-            }
-            if (!member.isSetter)
-              member.tpe match {
-                case MethodType(List(), ConstantType(_)) =>
-                  // member is a constant; only getter is needed
-                  ;
-                case _ =>
-                  // otherwise mixin a field as well
-                  addMember(clazz,
-                            clazz.newValue(member.pos, nme.getterToLocal(member.name))
-                            setFlag (LOCAL | PRIVATE | member.getFlag(MUTABLE | LAZY))
-                            setInfo member.tpe.resultType)
+          if (isConcreteAccessor(member)) {
+            if (isOverridden(member)) {
+              if (settings.debug.value) println("!!! is overridden val: "+member)
+            } else {
+              // mixin field accessors
+              val member1 = addMember(
+                clazz,
+                member.cloneSymbol(clazz)
+                  setPos clazz.pos
+                  setFlag FINAL resetFlag (DEFERRED | lateDEFERRED))
+              if (member.hasFlag(LAZY)) {
+                var init = implClass(mixinClass).info.decl(member.name)
+                assert(init != NoSymbol, "Could not find initializer for " + member.name)
+                initializer(member1) = init
               }
+              if (!member.isSetter)
+                member.tpe match {
+                  case MethodType(List(), ConstantType(_)) =>
+                    // member is a constant; only getter is needed
+                    ;
+                  case MethodType(List(), TypeRef(_, tpeSym, _))
+                    if member.hasFlag(LAZY) && tpeSym == definitions.UnitClass =>
+                    // member is a lazy value of type unit. No field needed
+                    ;
+                  case _ =>
+                    // otherwise mixin a field as well
+                    addMember(clazz,
+                              clazz.newValue(member.pos, nme.getterToLocal(member.name))
+                              setFlag (LOCAL | PRIVATE | member.getFlag(MUTABLE | LAZY))
+                              setFlag (if (!member.hasFlag(STABLE)) MUTABLE else 0)
+                              setInfo member.tpe.resultType
+                              setAttributes member.attributes)
+                }
+            }
           } else if (member hasFlag SUPERACCESSOR) { // mixin super accessors
             val member1 = addMember(clazz, member.cloneSymbol(clazz)) setPos clazz.pos
             assert(member1.alias != NoSymbol, member1)
@@ -352,7 +366,9 @@ abstract class Mixin extends InfoTransform {
      *   - For every static implementation method:
      *       - remove override flag
      *       - create a new method definition that also has a `self' parameter
-     *         (which comes first)
+     *         (which comes first) Iuli: this position is assumed by tail call elimination 
+     *         on a different receiver. Storing a new 'this' assumes it is located at
+     *         index 0 in the local variable table. See 'STORE_THIS' and GenJVM/GenMSIL.
      *   - Map implementation class types in type-apply's to their interfaces
      *   - Remove all fields in implementation classes
      */
@@ -475,7 +491,7 @@ abstract class Mixin extends InfoTransform {
         def isNotDuplicate(tree: Tree) = tree match {
           case DefDef(_, _, _, _, _, _) =>
             val sym = tree.symbol;
-            !((sym hasFlag DEFERRED) &&
+            !(sym.isDeferred &&
               (newSyms exists (nsym => nsym.name == sym.name && (nsym.tpe matches sym.tpe))))
           case _ =>
             true
@@ -531,7 +547,7 @@ abstract class Mixin extends InfoTransform {
        *  where bitmap$n is an int value acting as a bitmap of initialized values. It is
        *  the 'n' is (offset / 32), the MASK is (1 << (offset % 32)).
        */
-      def mkLazyDef(clazz: Symbol, init: Tree, fieldSym: Symbol, offset: Int): Tree = {
+      def mkLazyDef(clazz: Symbol, init: Tree, retVal: Tree, offset: Int): Tree = {
 
         /** Return the bitmap field for 'offset', create one if not inheriting it already. */ 
         def bitmapFor(offset: Int): Symbol = {
@@ -569,17 +585,21 @@ abstract class Mixin extends InfoTransform {
                         Literal(Constant(()))), 
                  EmptyTree)), 
              EmptyTree)
-        localTyper.typed(Block(List(result), Select(This(clazz), fieldSym)))
+        localTyper.typed(atPos(init.pos)(Block(List(result), retVal)))
       }
        
       /** Complete lazy field accessors. Applies only to classes, for it's own (non inherited) lazy fields. */
       def lazifyOwnFields(clazz: Symbol, stats: List[Tree]): List[Tree] = {
-        var offset = clazz.info.findMember(nme.ANYNAME, METHOD, LAZY, false).alternatives.filter(_.owner != clazz).length
+        var offset = clazz.info.findMember(nme.ANYNAME, 0, METHOD | LAZY, false).alternatives.filter(_.owner != clazz).length
         val stats1 = for (stat <- stats; sym = stat.symbol) yield stat match {
           case DefDef(mods, name, tp, vp, tpt, rhs) 
             if sym.hasFlag(LAZY) && rhs != EmptyTree && !clazz.isImplClass =>
-              val Block(List(assignment), res) = rhs
-              val rhs1 = mkLazyDef(clazz, assignment, res.symbol, offset)
+              val rhs1 = if (sym.tpe.resultType.typeSymbol == definitions.UnitClass) 
+                mkLazyDef(clazz, rhs, Literal(()), offset)
+              else {
+                val Block(List(assignment), res) = rhs
+                mkLazyDef(clazz, assignment, Select(This(clazz), res.symbol), offset)
+              }
               offset += 1
               copy.DefDef(stat, mods, name, tp, vp, tpt, rhs1)
           case _ => stat
@@ -589,7 +609,7 @@ abstract class Mixin extends InfoTransform {
        
       
       // the number of inherited lazy fields that are not mixed in
-      offset = (clazz.info.findMember(nme.ANYNAME, METHOD, LAZY, false)
+      offset = (clazz.info.findMember(nme.ANYNAME, 0, METHOD | LAZY, false)
                 .alternatives filter { f => f.owner != clazz || !f.hasFlag(MIXEDIN)}).length
       // begin addNewDefs
       var stats1 = lazifyOwnFields(clazz, stats)
@@ -600,8 +620,8 @@ abstract class Mixin extends InfoTransform {
           if (clazz hasFlag lateINTERFACE) {
             // if current class is a trait interface, add an abstract method for accessor `sym'
             addDefDef(sym, vparamss => EmptyTree)
-          } else if (!clazz.isTrait) {
-            // if class is not a trait:
+          } else if (!clazz.isTrait) { 
+            // if class is not a trait add accessor definitions
             if ((sym hasFlag ACCESSOR) && 
                 (!(sym hasFlag DEFERRED) || (sym hasFlag lateDEFERRED))) { 
               // add accessor definitions
@@ -611,11 +631,15 @@ abstract class Mixin extends InfoTransform {
                   case _ => 
                     // if it is a mixed-in lazy value, complete the accessor
                     if (sym.hasFlag(LAZY) && sym.isGetter) {
-                      val assign = 
-                        Assign(Select(This(sym.accessed.owner), sym.accessed) /*gen.mkAttributedRef(sym.accessed)*/, 
-                            Apply(staticRef(initializer(sym)), gen.mkAttributedThis(clazz) :: Nil))
-
-                      val rhs1 = mkLazyDef(clazz, assign, sym.accessed, offset)
+                      val rhs1 = if (sym.tpe.resultType.typeSymbol == definitions.UnitClass)
+                        mkLazyDef(clazz, Apply(staticRef(initializer(sym)), List(gen.mkAttributedThis(clazz))), Literal(()), offset)
+                      else {
+                        val assign = atPos(sym.pos) {
+                          Assign(Select(This(sym.accessed.owner), sym.accessed) /*gen.mkAttributedRef(sym.accessed)*/, 
+                              Apply(staticRef(initializer(sym)), gen.mkAttributedThis(clazz) :: Nil))
+                        }
+                        mkLazyDef(clazz, assign, Select(This(clazz), sym.accessed), offset)
+                      }
                       offset += 1 
                       rhs1
                     } else
@@ -628,7 +652,7 @@ abstract class Mixin extends InfoTransform {
               // add modules 
               val vdef = gen.mkModuleVarDef(sym)
               addDef(position(sym), vdef)
-              addDef(position(sym), gen.mkModuleAccessDef(sym, vdef.symbol))
+              addDef(position(sym), gen.mkCachedModuleAccessDef(sym, vdef.symbol))
             } else if (!sym.isMethod) {
               // add fields
               addDef(position(sym), ValDef(sym))
